/**************************************************************
*
* emtaglenoise.cc	XMM Science Analysis System (SAS)
*
* Task interface to the AnomalousStateFinder class of routines.
*
* (c) 2008 ESA - Written by Hubert Chen, CEA Saclay
*
* $Id: emtaglenoise.cc,v 0.6 2009/04/15 10:02:28 cmhchen Exp $
*
**************************************************************/

// C++ global:
#include "emtaglenoise.h"
	// automatically generated; main() needs this.
// Implementation specific:
#include "AnomalousStateFinder.h"
#include <evselect.h>

/** Given the filename of a MOS merged event list (on the command line),
    decide if any CCD is in an anomalous state, and write the result
    as attributes to the input file header.
    This is the SAS task that gives access to the AnomalousStateFinder class. */
void emtaglenoise() {
//  AnomalousStateFinder::tagAnomalous( stringParameter("eventset") );
// It gets more involved now that we need to know which CCD(s) are tagged
//  in order to filter.

  string eventset = stringParameter("eventset");
  AnomalousStateFinder * anomalousStateFinderPtr
	= new AnomalousStateFinder(eventset);
  // Compute and write the attributes.
  anomalousStateFinderPtr->writeAttributes();

  // Filter DataSet if desired.
  string filterExpression;
  if ( booleanParameter("filterbadccds") ) {
    // First find out which CCD(s) are tagged.
    ostringstream filterExpressionStream;
    filterExpressionStream << "true";
    for (Ccd ccdnr=EMOScam::CCD_FIRST; ccdnr <= EMOScam::CCD_LAST; ccdnr++)
      if ( anomalousStateFinderPtr->isAnomalous(ccdnr) )
	filterExpressionStream << " && (CCDNR != " << ccdnr << ")";
    filterExpression = filterExpressionStream.str();
  } // if booleanParameter("filterbadccds")

  // Important: Must delete the AnomalousStateFinder _before_ executing
  //  evselect, or else the new Attributes will not have been written
  //  to eventset, and the filteredset will thus not contain the attributes.
  delete anomalousStateFinderPtr; anomalousStateFinderPtr = 0;

  if ( booleanParameter("filterbadccds") ) {
    // Call evselect to filter.
    Info::Task * evselect = task("evselect");
    if (evselect == NULL)
      Sas::error("Task not found", "Failed to call evselect for filtering.");
    evselect->setParameter( "table"      , eventset );
    evselect->setParameter( "filteredset", stringParameter("filteredset") );
    evselect->setParameter( "expression" , filterExpression );
    evselect->setParameter( "writedss"   , true );
    evselect->exec();
  } // if booleanParameter("filterbadccds")
} // void emtaglenoise()

/*
$Log: emtaglenoise.cc,v $
Revision 0.6  2009/04/15 10:02:28  cmhchen
Renamed package from goodpixelcount to emtaglenoise.
Renamed task    from emtaganomalous to emtaglenoise.
emtaglenoise now writes LENOISnn to file instead of printing to STDERR.
Implemented the filterbadccds and filteredset parameters.

Revision 0.5  2008/10/28 14:40:21  cmhchen
Inserted '#include "emtaganomalous.h"' previously in AnomalousStateFinder.h .

Revision 0.4  2008/10/24 17:49:01  cmhchen
Initial revision.

*/
