/********************************************************
*
* goodpixelcount.cc	XMM Science Analysis System (SAS)
*
* Class containing routines to calculate and write no. of
* good and selected MOS pixels to DataSets.
*
* (c) 2008 ESA - Written by Hubert Chen, CEA Saclay
*
* $Id: PixelCounter.cc,v 0.4 2008/10/25 23:02:13 cmhchen Exp $
*
********************************************************/

#include "PixelCounter.h"

// Potential problem: Is int 16- or 32-bit?  Can we be sure that it is 32?
// No. good pixels can easily be > 2^15.

// To do:
// Read in DSS extension and do spatial region filtering.
// Requires: Converting (RAWX,RAWY) to (DETX,DETY).
// Strategy:
// in write...(), call new DSS filtering routines.
// in write, if filtering expression is given, call new filtering routine.

// DSS filtering routine should check DataSet for presence of DSS extension.
// If not, return.  If yes, call DataSubspace::selectExpression(???) to get
//  string filter expression, then call expression filtering routine.

// string filtering routine should...

// need to modify function to not return raw pixel counts, but to return
// boolean raw pixel array, and new routine to convert array to count.

// need routine to convert RAWX/Y PIXCOORD1 to CHIPCOORD to CAMCOORD1
//  to CAMCOORD2 to DETX/Y (Look in cal).

// routine to apply filters to arrays.  (selectlib?)

//------------------------------------------------------------------------------
PixelCounter::PixelCounter(const string filename) {
// filename |-> DataSet.
// Call the constructor that takes *DataSet as input.
  // Open file.
  _dataSetPtr = dataSetServer->open( filename, Dal::Read );
  init();
} // PixelCounter::PixelCounter(const string filename)

//------------------------------------------------------------------------------
PixelCounter::PixelCounter(DataSet * dataSetPtr) {
  _dataSetPtr = dataSetPtr;
  init();
} // PixelCounter::PixelCounter(const string filename)

//------------------------------------------------------------------------------
void PixelCounter::init() {
errstr << "init() " << Sas::message(Msg::App,Msg::Always);
/*
  _expr = _dataSetPtr->hasBlock("EVENTS")?
	DataSubspace(_dataSetPtr->block("EVENTS")).selectExpression()
	: "";
  if ( _dataSetPtr->hasBlock("EVENTS") ) {
    Block * eventBlockPtr = _dataSetPtr->block("EVENTS");
    _expr = DataSubspace(eventBlockPtr).selectExpression();
  } // if ( _dataSetPtr->hasBlock("EVENTS") )
errstr << "_expr=" << _expr << Sas::message(Msg::App,Msg::Always);
// Expressions returned by DataSubspace.selectExpression() turn out
//  to be too complicated to decipher, plus, it includes numerous
//  variables besides spatial coordinates.
*/
  _expr = "";

  // Set the state of the calServer.
  // Must do this for calServer.instrument().* to function properly.
  calServer.setState(*_dataSetPtr);
  // We do not want randomization when converting RAWX/Y to DETX/Y.
  calServer.getState().randomize()->set(false);
//  if ( _dataSetPtr->hasBlock("EVENTS") )
//    calServer.setState( *_dataSetPtr->block("EVENTS") );
//  else calServer.setState(*_dataSetPtr);
//  if ( _dataSetPtr->hasBlock("BADPIX02") )
//    calServer.setState( *_dataSetPtr->block("BADPIX03") );
//    calServer.getState().ccd()->set(4);
//  calServer.setCcf  (*_dataSetPtr);	// must get filename...
  const XmmInstrument * instrument = calServer.instrument();
//errstr << "xmminstrument.id()="<<instrument->id()<<Sas::message(Msg::App,Msg::Always);

  // Initialize state variables.
  _rawXmin = instrument->minX();
  _rawXmax = instrument->maxX();
  _rawYmin = instrument->minY();
  _rawYmax = instrument->maxY();
  _nCcds   = instrument->nCcds();
  _status = new bool[_nCcds];
  _goodPixelCounts = new long[_nCcds];
  _goodPixelMaps   = new (bool **)[_nCcds];
  _selectedPixelCounts = new long[_nCcds];
  _selectionValidity = new bool[_nCcds];
  for (unsigned i=0u, j=EMOScam::CCD_FIRST; j<=EMOScam::CCD_LAST; i++, j++) {
    _status[i] = false;
    _goodPixelCounts[i] = -1L;
  } // for (unsigned i=0u, j=EMOScam::CCD_FIRST; j>EMOScam::CCD_LAST; i++, j++)
  vector<unsigned long> dummy = goodRawPixelCounts();
	// This call ensures that _goodPixelMaps[][][] is made valid.
  assert(_goodPixelMaps[0] != 0);
//errstr << _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
//errstr << _goodPixelMaps[1] << Sas::message(Msg::App,Msg::Always);
errstr<<"state="<<calServer.getState()<<Sas::message(Msg::App,Msg::Always);
} // void PixelCounter::init()

//------------------------------------------------------------------------------
PixelCounter::~PixelCounter() {
  // Close file.
//  dataSetServer->close(_dataSetPtr);

//errstr << "goodpixelcounts:";
//for (unsigned i=0; i < _nCcds; i++) errstr<<" "<<_goodPixelCounts[i];
//errstr << Sas::message(Msg::App,Msg::Always);
  // Delete state variables
  delete [] _status; _status = 0;
  delete [] _goodPixelCounts; _goodPixelCounts = 0;

  unsigned nX = _rawXmax - _rawXmin + 1u;
// I do not know what is wrong with the following.
// _goodPixelMaps, _goodPixelMaps[ccdnr] and _goodPixelMaps[ccdnr][i] are
// valid but accessing _goodPixelMaps[ccdnr][i][j] produces segfault.
//if (_goodPixelMaps != NULL) {
//cout << _goodPixelMaps << endl;
  for (unsigned ccdnr=0u; ccdnr < _nCcds; ccdnr++) {
//if (_goodPixelMaps[ccdnr] != 0)
//{
//cout << _goodPixelMaps[ccdnr] << endl;
    for (unsigned i=0u; i < nX; i++)
//;
if (_goodPixelMaps[ccdnr][i] != NULL)
{
//cout << _goodPixelMaps[ccdnr][i];
for (unsigned j=0; j <= _rawYmax - _rawYmin; j++) {
//cout << _goodPixelMaps[ccdnr][i][j];
//cout << endl;
      delete [] _goodPixelMaps[ccdnr][i]; _goodPixelMaps[ccdnr][i] = 0;
} // for (unsigned j=0; j <= _rawYmax - _rawYmin; j++)
//}
//cout << endl;
    delete [] _goodPixelMaps[ccdnr]; _goodPixelMaps[ccdnr] = 0;
}
  } // for (int ccdnr=0; ccdnr < _nCcds; ccdnr++)
//}
  delete [] _goodPixelMaps; _goodPixelMaps = 0;
errstr << "destructor done" << Sas::message(Msg::App,Msg::Always);
} // PixelCounter::~PixelCounter()

//------------------------------------------------------------------------------
// Given the filename of an event list (on the command line),
// compute the number of good pixels on each CCD, based on the content
// of the BADPIXnn extensions, and if a DSS extension is present,
// include only pixels within the selection.  Write the pixel counts
// to the file header as values of keyword attributes NGOODPIn (n=1..7)
// for each CCD and NGOODPIX for their sum.
// This is an XMMSAS task that reads parameters via the common mechanism.
void PixelCounter::writeGoodPixelCount(const string eventSetName) {
  PixelCounter pixelCounter(eventSetName);
  vector<unsigned long> counts = pixelCounter.goodPixelCounts();

  // Calculate and write nGoodPix to file as keywords.
  // NGOODPIn, n=EMOScam::CCD_FIRST..EMOScam::CCD_LAST are
  //  the no. good pixels on each CCD; NGOODPIX is their sum.
  DataSet * eventSetPtr = dataSetServer->open(eventSetName, Dal::Modify);
  unsigned long nGoodPixAll = 0L;
  for (Ccd ccdnr=EMOScam::CCD_FIRST; ccdnr<=EMOScam::CCD_LAST; ccdnr++) {
//    unsigned long nGoodPixI = pixelCounter.goodRawPixelCount(ccdnr);
    unsigned long nGoodPixI = counts[ccdnr - EMOScam::CCD_FIRST];
    ostringstream ccdnrStrStream;
    ccdnrStrStream << ccdnr;
    const string ccdnrStr = ccdnrStrStream.str();
    Attribute * NGOODPIn
	= eventSetPtr->addAttribute("NGOODPI"+ccdnrStr,
			"No. good, selected pixels in MOS CCD "+ccdnrStr);
    *NGOODPIn = (int)nGoodPixI;
    nGoodPixAll += nGoodPixI;
  } // for (Ccd ccdnr=EMOScam::CCD_FIRST; ccdnr<=EMOScam::CCD_LAST; ccdnr++)
  Attribute * NGOODPIX
    = eventSetPtr->addAttribute("NGOODPIX",
				"No. good, selected pixels in all MOS CCDs");
  *NGOODPIX = (int)nGoodPixAll;

//  delete pixelCounter;
} // void PixelCounter::writeGoodPixelCount(const string eventSetName)

//------------------------------------------------------------------------------
void PixelCounter::printGoodPixelCount(const string eventSetName,
					const string expr,
					const bool cornerFlag) {
//  PixelCounter pixelCounter(eventSetName);
  PixelCounter pixelCounter = PixelCounter(eventSetName);
  pixelCounter.addExpression(expr);
  // Add the `corner pixel' criteria to the existing expression, if any.
  if (cornerFlag)
    pixelCounter.addExpression( cornerExpression(calServer.instrument()->id()) );
  vector<unsigned long> counts = pixelCounter.goodPixelCounts();
  errstr << "Good MOS pixel counts =";
//  for (vector<unsigned long>::iterator countsIterator = counts.begin();
//	countsIterator != counts.end(); countsIterator++)
//    errstr << (unsigned long)countsIterator;
  for (unsigned i=0u; i < counts.size(); i++)
    errstr << " " << counts[i];
  errstr << Sas::message(Msg::App,Msg::Always);
//  delete pixelCounter;
} // void PixelCounter::writeGoodPixelCount(const string eventSetName)

//------------------------------------------------------------------------------
vector<unsigned long> PixelCounter::goodPixelCounts() {
//errstr << "goodPixelCounts()" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  vector<unsigned long> output(_nCcds);
  Ccd ccdnr = EMOScam::CCD_FIRST;
  for (unsigned i = 0u; i < _nCcds; i++, ccdnr++)
    output[i] = goodPixelCount(ccdnr);
  return output;
} // vector<unsigned long> PixelCounter::goodPixelCounts()

//------------------------------------------------------------------------------
unsigned long PixelCounter::goodPixelCount(Ccd ccdnr) {
//errstr << "goodPixelCount(" << ccdnr<< ")" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
/*
  unsigned arrayIndex = ccdnr - EMOScam::CCD_FIRST;
  if ( _status(arrayIndex) )? _goodPixelCounts[arrayIndex]
  return ( _status(arrayIndex) )? _goodPixelCounts[arrayIndex]
*/
//  return goodRawPixelCount(ccdnr);
  return filterCCD(ccdnr);
} // unsigned long PixelCounter::goodPixelCount(Ccd ccdnr)

//------------------------------------------------------------------------------
void PixelCounter::addExpression(const string expr) {
  _expr = (_expr == "") ? expr : "("+_expr+") && ("+expr+")";
  for (unsigned i=0u; i < _nCcds; i++)
    _selectionValidity[i] = false;
  errstr << "New selection expression:" << _expr
	<< Sas::message(Msg::App,Msg::Verbose);
errstr << "New selection expression:" << _expr << Sas::message(Msg::App,Msg::Always);
} // void PixelCounter::addExpression(const string expr)

//------------------------------------------------------------------------------
const string PixelCounter::cornerExpression(Instrument instrument) {
  if (instrument == Xmm::EMOS1)
    return EMOS1cornerExpression;
  if (instrument == Xmm::EMOS2)
    return EMOS2cornerExpression;
  errstr << "Instrument type " << instrument << " not recognised."
	<< Sas::error("Unrecognised instrument.");
  return "";	// Execution will never reach this point.
} // const string PixelCounter::cornerExpression(Instrument instrument)

//------------------------------------------------------------------------------
// This constant string should really be in a header file (or a config file?)
// somewhere, but initialization within class definition is forbidden...
const string PixelCounter::EMOS1cornerExpression =
	"!("
	    "circle(100, -200, 17700, DETX, DETY)"
	" || circle(834,  135, 17100, DETX, DETY)"
	" || circle(700, -803, 17100, DETX, DETY)"
	" || box(-   20, -17000, 6500,  500,   0, DETX, DETY)"
	" || box(  5880, -20500, 7500, 1500,  10, DETX, DETY)"
	" || box(- 5920, -20500, 7500, 1500, 350, DETX, DETY)"
	" || box(-   20, -20000, 5500,  500,   0, DETX, DETY)"
	" || box(-12900,  16000,  250, 4000,   0, DETX, DETY)"
	" || box(    80,  18500,  150, 1300,   0, DETX, DETY)"
	" || box(-   10, -18800,  125, 1500,   0, DETX, DETY)"
	")";

//------------------------------------------------------------------------------
// This constant string should really be in a header file (or a config file?)
// somewhere, but initialization within class definition is forbidden...
const string PixelCounter::EMOS2cornerExpression =
	"!("
	    "circle( 435, 1006, 17100, DETX, DETY)"
	" || circle(- 34,   68, 17700, DETX, DETY)"
	" || box(-  20, -17000, 6500,  500,   0, DETX, DETY)"
	" || box( 5880, -20500, 7500, 1500,  10, DETX, DETY)"
	" || box(-5920, -20500, 7500, 1500, 350, DETX, DETY)"
	" || box(-  20, -20000, 5500,  500,   0, DETX, DETY)"
	")";

//------------------------------------------------------------------------------
vector<unsigned long> PixelCounter::goodRawPixelCounts() {
  vector<unsigned long> goodRawPixelCounts = vector<unsigned long>(_nCcds, 0ul);
  for (unsigned i=0u; i < _nCcds; i++)
    goodRawPixelCounts[i] = goodRawPixelCount(i + EMOScam::CCD_FIRST);
  return goodRawPixelCounts;
} // vector<unsigned long> PixelCounter::goodRawPixelCounts()

//------------------------------------------------------------------------------
unsigned long PixelCounter::goodRawPixelCount(const Ccd ccdnr) {
  unsigned arrayIndex = ccdnr - EMOScam::CCD_FIRST;
  if (_goodPixelCounts[arrayIndex] < 0)
    _goodPixelCounts[arrayIndex] = sumGoodPixelMap(ccdnr);
  return _goodPixelCounts[arrayIndex];
} // bool[][] PixelCounter::goodRawPixelCount(const Ccd ccdnr)

//------------------------------------------------------------------------------
unsigned long PixelCounter::sumGoodPixelMap(const Ccd ccdnr) {
  unsigned arrayIndex = ccdnr - EMOScam::CCD_FIRST;
errstr << "sumGoodPixelMap("<<ccdnr<<")" << Sas::message(Msg::App,Msg::Always);
//  if (!_status[arrayIndex])
  if (_goodPixelCounts[arrayIndex] < 0)
{
//errstr << "_goodPixelCounts < 0 ccdnr=" << ccdnr << Sas::message(Msg::App,Msg::Always);
    assert(_goodPixelMaps != 0);
    _goodPixelMaps[arrayIndex] = computeGoodPixelMap(ccdnr);
}
    assert(_goodPixelMaps[arrayIndex] != 0);
    assert(_goodPixelMaps != 0);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
    assert(_goodPixelMaps[0] != 0);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
//errstr << "_goodPixelCounts[0] ="<< _goodPixelCounts[0] << Sas::message(Msg::App,Msg::Always);
//errstr << "bool*  ="<< _goodPixelMaps[0][0] << Sas::message(Msg::App,Msg::Always);

  unsigned long nX = _rawXmax - _rawXmin + 1ul;
  unsigned long nY = _rawYmax - _rawYmin + 1ul;
  // Return the number of good pixels within selected region.
  unsigned long myGoodPixelCount = 0ul;
  for (unsigned i=0u; i<nX; i++)
    for (unsigned j=0u; j<nY; j++)
{
    assert(_goodPixelMaps[arrayIndex] != 0);
    assert(_goodPixelMaps[arrayIndex][i] != 0);
      if (_goodPixelMaps[arrayIndex][i][j])
	myGoodPixelCount++;
}
//else if (ccdnr == 1) errstr << "x=" << i << " y=" << j << Sas::message(Msg::App,Msg::Always);
  return myGoodPixelCount;
} // unsigned long PixelCounter::sumGoodPixelMap(const Ccd ccdnr)

//------------------------------------------------------------------------------
bool *** PixelCounter::getGoodPixelMaps() {
  for (unsigned i=0u; i < _nCcds; i++)
//    if (!_status[i])
    if (_goodPixelCounts[i] < 0)
      _goodPixelMaps[i] = computeGoodPixelMap(i + EMOScam::CCD_FIRST);
  return _goodPixelMaps;
} // bool *** PixelCounter::getGoodPixelMaps()

//------------------------------------------------------------------------------
bool ** PixelCounter::getGoodPixelMap(const Ccd ccdnr) {
  unsigned arrayIndex = ccdnr - EMOScam::CCD_FIRST;
//  if (!_status[arrayIndex])
  if (_goodPixelCounts[arrayIndex] < 0)
    _goodPixelMaps[arrayIndex] = computeGoodPixelMap(ccdnr);
  return _goodPixelMaps[arrayIndex];
} // bool ** PixelCounter::getGoodPixelMap(const Ccd ccdnr)

//------------------------------------------------------------------------------
// Given a DataSet pointer and a Ccd ccdnr (see caloalutils/src/Xmm.h),
// read the extension BADPIX0n (n=CCDNR) from the DataSet,
// compute and return the number of good pixels on the given CCD,
// within the selected region if a DSS extension is present.
//
// Caution: The state variables _raw{X,Y}{min,max} must be assigned
//  values (usually in the constructor) prior to calling this function.
bool ** PixelCounter::computeGoodPixelMap(const Ccd ccdnr) {
errstr << "computeGoodPixelMap " << ccdnr << Sas::message(Msg::App,Msg::Always);
  // Make event list that covers an entire detector.
  const unsigned nX = _rawXmax - _rawXmin + 1u;
  const unsigned nY = _rawYmax - _rawYmin + 1u;
//  bool rawPixelFlags[nX][nY];
  bool ** rawPixelFlags = new (bool*)[nX];
  assert(rawPixelFlags != 0);
  for (unsigned i=0u; i<nX; i++) {
    rawPixelFlags[i] = new bool[nY];
    for (unsigned j=0u; j<nY; j++)
{
  assert(rawPixelFlags != 0);
      rawPixelFlags[i][j] = true;
}
  } // for (unsigned i=0u; i<nX; i++)

  // Read the content of BADPIXnn extensions from dataSet, if present.
  ostringstream blockNameStream;
  blockNameStream << "BADPIX0" << ccdnr;
  const string blockName = blockNameStream.str();
  if ( _dataSetPtr->hasBlock(blockName) ) {
    // BADPIXnn extensions are BINTABLEs with the following columns:
    // RAWX, RAWY, TYPE, YEXTENT and BADFLAG
    const Table * badPixTable = _dataSetPtr->table(blockName, Dal::Read);
    const ColumnData * rawXdata = badPixTable->column("RAWX", Dal::Read)->data();
    const ColumnData * rawYdata = badPixTable->column("RAWY", Dal::Read)->data();
    const ColumnData * type = badPixTable->column("TYPE", Dal::Read)->data();
    const ColumnData * yExtentData = badPixTable->column("YEXTENT", Dal::Read)->data();
    const ColumnData * badFlag = badPixTable->column("BADFLAG", Dal::Read)->data();
    const int16 * rawX = rawXdata->int16Data();
    const int16 * rawY = rawYdata->int16Data();
    const int16 * yExtent = yExtentData->int16Data();

    const unsigned long Nrows = badPixTable->column("RAWX", Dal::Read)->rows();
    // Null the flags of bad pixels.
    for (unsigned long i=0ul; i < Nrows; i++) {
//if (ccdnr == 1)
//errstr << "x=" << rawX[i] << " y=" << rawY[i] << " n_elements=" << Nrows << Sas::message(Msg::App,Msg::Always);
//errstr << "computeGoodPixelMap " << "b" << Sas::message(Msg::App,Msg::Always);
      if ( (unsigned)(rawY[i] + yExtent[i] - 1) > _rawYmax )
	Sas::error("invalidIndex", "RAWY+YEXTENT-1 is out of range.");
//errstr << "computeGoodPixelMap " << "c" << Sas::message(Msg::App,Msg::Always);
      for (int yOffset=0; yOffset < yExtent[i]; yOffset++)
//{ errstr << "rawX[i]=" << rawX[i] << " _rawXmin=" << _rawXmin << " rawY[i]=" << rawY[i] << " _rawYmin=" << _rawYmin << " yOffset=" << yOffset << Sas::message(Msg::App,Msg::Always);
//{errstr<<"("<<ccdnr<<","<<rawX[i]<<","<<rawY[i]<<") is bad."<<Sas::message(Msg::App,Msg::Always);
{ assert(rawPixelFlags != 0);
	rawPixelFlags	[ rawX[i]-_rawXmin ]
			[ rawY[i]-_rawYmin+yOffset ] = false;
}
//}
    } // for (unsigned long i=0ul; i < Nrows; i++)

    // ColumnData instances must be deleted to avoid memory leak.
    // See the C++ API for Dal, under CLASS Column.
    delete rawXdata; rawXdata = 0;
    delete rawYdata; rawYdata = 0;
    delete type; type = 0;
    delete yExtentData; yExtentData = 0;
    delete badFlag; badFlag = 0;
  } // if _dataSetPtr->hasBlock(blockName)
  else {
    // No BADPIXnn extension.  Do we warn the caller or just do nothing?
  } // if ! _dataSetPtr->hasBlock(blockName)

  _status[ccdnr - EMOScam::CCD_FIRST] = true;
  return rawPixelFlags;
} // bool ** PixelCounter::computeGoodPixelMap(const Ccd ccdnr)

//------------------------------------------------------------------------------
/** wrapper around filterCCD{fine,coarse}(Ccd ccdnr) */
unsigned long PixelCounter::filterCCD(Ccd ccdnr, bool fine_flag) {
  return (fine_flag)? filterCCDfine(ccdnr) : filterCCDcoarse(ccdnr);
} // unsigned long PixelCounter::filterCCD(Ccd ccdnr, bool fine_flag)

//------------------------------------------------------------------------------
/** Compute the number of good and selected RAWX/Y pixels on CCD ccdnr. */
unsigned long PixelCounter::filterCCDcoarse(Ccd ccdnr) {
errstr << "filterCCDcoarse" << Sas::message(Msg::App,Msg::Always);

  // Preparations for filtering with input expression.
  Evaluator evaluator(_expr);
  NumericSymbols syms = evaluator.numericIdentifiers();
  int detXptrPos = -1;
  int detYptrPos = -1;
  // check syms and match fields with (x,y)
  for (unsigned i=0u; i < syms.name.size(); i++)
    if (syms.name[i] == "DETX")
      detXptrPos = i;
    else
    if (syms.name[i] == "DETY")
      detYptrPos = i;
    else
      Sas::error("invalid selection expression",
	"Selection expression cannot contain variables other than DETX/Y.");

  // Preparations for coordinate transformation (RAWX/Y |-> DETX/Y).
	// Must set ccdnr before calling calServer.getAtom(GeometryDataServer);
//errstr<<"randomize="<<calServer.getState().randomize()->get()<< Sas::message(Msg::App,Msg::Always);
  PositionUtils * positionUtils
	= new PositionUtils( calServer.instrument()->id() );
	// PositionUtils() resets the CCDNR to 1.
  calServer.getState().ccd()->set(ccdnr);
/*
  GeometryDataServer * gds = 0; gds = calServer.getAtom(gds);
      MiscDataServer * mds = 0; mds = calServer.getAtom(mds);
  double mmX_to_detX =	  mds->get( "PLATE_SCALE_X") * 20.0
			/ mds->get("MM_PER_PIXEL_X");
  double mmY_to_detY =	  mds->get( "PLATE_SCALE_Y") * 20.0
			/ mds->get("MM_PER_PIXEL_Y");
	// Reference: packages/ccftools/coordconv/src/camtochip.cc
  double mmX_to_detX = (180. * 3600 * 20) / mds->get("FOCAL_LENGTH") / M_PI;
  double mmY_to_detY = (180. * 3600 * 20) / mds->get("FOCAL_LENGTH") / M_PI;
	// Reference: packages/arfgen/src/PositionUtils.cc
  calServer.releaseAtom(mds);
*/

  vector<unsigned long> dummy = goodRawPixelCounts();
	// This call ensures that _goodPixelMaps[][][] is made valid.
  assert(_goodPixelMaps[0] != 0);

  unsigned long goodPixelCount = 0ul;
  unsigned arrayIndex = ccdnr - EMOScam::CCD_FIRST;
  unsigned long nX = _rawXmax - _rawXmin + 1ul;
  unsigned long nY = _rawYmax - _rawYmin + 1ul;
  // Loop over integral RAWX/Y.
  for (unsigned i=0u; i < nX; i++)
    for (unsigned j=0u; j < nY; j++) {
      // Skip the bad pixels.
      if (!_goodPixelMaps[arrayIndex][i][j])
	continue; // for (unsigned j=0u; j < nY; j++)

      // Coordinate transformation: RAWX/Y |-> DETX/Y
/*
      int16 rawX = i;
      int16 rawY = j;
      CalInt16Vector rawXvector(&rawX, 1u);
      CalInt16Vector rawYvector(&rawY, 1u);
      XmmCartesianPointList camCoord2List
	= gds->pixCoord1List(rawXvector, rawYvector)
	.toChipCoord()
	.toCamCoord1()	// The correctness of this call requires
			// calServer.getState().ccd()->set(ccdnr);
	.toCamCoord2();
      CalReal64Vector mmX = camCoord2List.x();
      CalReal64Vector mmY = camCoord2List.y();
      double detX = mmX[0] * mmX_to_detX;
      double detY = mmY[0] * mmY_to_detY;
*/
      CoordPair detXY = positionUtils->RawTodetxy005((double)i, (double)j);
	// CoordPair declared in datautils/src/CoordPair.h
//errstr<<"arfgen:"<<detXY.x()<<","<<detXY.y()<<" platescale:"<<detX<<","<<detY<<Sas::message(Msg::App,Msg::Always);
//errstr<<detXY.x()<<" "<<detXY.y()<<" "<<detX<<" "<<detY<<Sas::message(Msg::App,Msg::Always);

      // Filter with DSS and/or selectlib expression.
/*
      if (detXptrPos >= 0) *(syms.valuePtr[detXptrPos]) = detX;
      if (detYptrPos >= 0) *(syms.valuePtr[detYptrPos]) = detY;
*/
      if (detXptrPos >= 0) *(syms.valuePtr[detXptrPos]) = detXY.x();
      if (detYptrPos >= 0) *(syms.valuePtr[detYptrPos]) = detXY.y();
      if ( !evaluator.evalLogic() )
	continue; // for (unsigned j=0u; j < nY; j++)

      // Passed all tests.  Remember the good pixel.
      goodPixelCount++;
    } // for (unsigned j=0u; j < nY; j++)

//  calServer.releaseAtom(gds);
  delete positionUtils; positionUtils = 0;

  // Multiply the RAWX/Y count by the RAWX/Y pixel area in DETX/Y units.
  vector<double> rawXYpixelwidth = PositionUtils( calServer.instrument()->id() )
					.detxyWidth();
	// From my reading of arfgen/src/PositionUtils.cc, version 1.76,
	// PositionUtils.detxyWidth() returns a 2-element vector of doubles
	// containing:
	//	{ detX( rawXYpixel(11,11) ) - detX( rawXYpixel(10,10) ),
	//	  detY( rawXYpixel(11,11) ) - detY( rawXYpixel(10,10) ) }.
	// It assumes calServer.state()->ccd()->set(ccdnr); has been called.
errstr<<"detxywidth("<<ccdnr<<")="<<rawXYpixelwidth[0]<<","<<rawXYpixelwidth[1]<<Sas::message(Msg::App,Msg::Always);
  return (unsigned long)(rawXYpixelwidth[0]
			*rawXYpixelwidth[1]
			*goodPixelCount    );
} // unsigned long PixelCounter::filterCCDcoarse(Ccd ccdnr)

//------------------------------------------------------------------------------
/** Compute the number of good and selected DETX/Y pixels on CCD ccdnr.
	Caution: This routine runs *forever*. */
unsigned long PixelCounter::filterCCDfine(Ccd ccdnr) {
errstr<<"filterCCD " << ccdnr << Sas::message(Msg::App,Msg::Always);
//errstr<<"filterCCD " << ccdnr << "(1)" << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  unsigned arrayIndex = ccdnr - EMOScam::CCD_FIRST;
//errstr<<"filterCCD " << ccdnr << "(2)" << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  if (!_selectionValidity[arrayIndex]) {
//errstr<<"filterCCD " << ccdnr << "(3)" << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
    list<scanningLib::Vertex> ccdCorners = CCDcorners(ccdnr);
//errstr<<"filterCCD " << ccdnr << "(4)" << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
    scanningLib::PolygonScanner polygonScanner(ccdCorners);
//errstr<<"filterCCD " << ccdnr << "(5)" << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
    calServer.getState().ccd()->set(ccdnr);
	// Must set ccdnr before calling filterRegion(PolygonScanner).
    _selectedPixelCounts[arrayIndex] = filterRegion(polygonScanner);
    _selectionValidity[arrayIndex] = true;
  } // if (!_selectionValidity[arrayIndex])
  return _selectedPixelCounts[arrayIndex];
} // unsigned long PixelCounter::filterCCDfine(Ccd ccdnr)

//------------------------------------------------------------------------------
/*
unsigned long PixelCounter::filterRegion(GeoRegion region) {
  filterRegion( scanningLib::RegionScanner(region,
					scanningLib::Grid(),
					scanningLib::Grid() ) );
} // unsigned long PixelCounter::filterRegion(GeoRegion region)
*/

//------------------------------------------------------------------------------
unsigned long PixelCounter::filterRegion(scanningLib::PolygonScanner shapeScanner) {
// Only deal with regions within a single CCD.
// Must set ccdnr within calServer.getState() before calling this function.
errstr << "filterRegion" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  Evaluator evaluator(_expr);
  NumericSymbols syms = evaluator.numericIdentifiers();
  int detXptrPos = -1;
  int detYptrPos = -1;
  // check syms and match fields with (x,y)
  for (unsigned i=0u; i < syms.name.size(); i++)
    if (syms.name[i] == "DETX")
      detXptrPos = i;
    else
    if (syms.name[i] == "DETY")
      detYptrPos = i;
    else
      Sas::error("invalid selection expression",
	"Selection expression cannot contain variables other than DETX/Y.");

  GeometryDataServer * gds = 0; gds = calServer.getAtom(gds);
      MiscDataServer * mds = 0; mds = calServer.getAtom(mds);
  double detX_to_mmX =	  mds->get("MM_PER_PIXEL_X") * 0.05
			/ mds->get( "PLATE_SCALE_X");
  double detY_to_mmY =	  mds->get("MM_PER_PIXEL_Y") * 0.05
			/ mds->get( "PLATE_SCALE_Y");
	// Reference: packages/ccftools/coordconv/src/camtochip.cc
  calServer.releaseAtom(mds);
//errstr << "filterRegion" << Sas::message(Msg::App,Msg::Always);
  vector<unsigned long> dummy = goodRawPixelCounts();
	// This call ensures that _goodPixelMaps[][][] is made valid.
  assert(_goodPixelMaps[0] != 0);

  unsigned long goodPixelCount = 0ul;
  shapeScanner.reset();
  list<scanningLib::Span> spanlist;
  int columnID;
errstr << "filterRegion: while..." << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  while ( shapeScanner.next(columnID, spanlist) ) {
    // Scan pixels.
    for (list<scanningLib::Span>::iterator spanIterator = spanlist.begin();
	spanIterator != spanlist.end(); spanIterator++) {
      scanningLib::Span span = *spanIterator;
      for (int rowID = span.min; rowID <= span.max; rowID++) {
	if (detXptrPos >= 0) *(syms.valuePtr[detXptrPos]) =    rowID;
	if (detYptrPos >= 0) *(syms.valuePtr[detYptrPos]) = columnID;

	// Filter with DSS and/or selectlib expression.
//errstr << "filterRegion: evaluate..." << Sas::message(Msg::App,Msg::Always);
	if ( !evaluator.evalLogic() )
//{errstr<<"("<<columnID<<","<<rowID<<") rejected."<<Sas::message(Msg::App,Msg::Always);
	  continue; // for (int rowID = span.min; rowID <= span.max; rowID++)
//}
//errstr << "filterRegion: evaluator done..." << Sas::message(Msg::App,Msg::Always);

	double mmX = (double)columnID * detX_to_mmX;
	CalReal64Vector xmmVector(&mmX, 1u);
	double mmY = (double)   rowID * detY_to_mmY;
	CalReal64Vector ymmVector(&mmY, 1u);
	double mmZ = 0.0;	// according to SAS documentation on CAL
	CalReal64Vector zmmVector(&mmZ, 1u);
	XmmCartesianPointList detXY
		= gds->camCoord2List(xmmVector, ymmVector, zmmVector);

	// Check for out-of-bound coordinates.
//	Ccd ccdnr = detXY.onCcd();
	Ccd ccdnr = gds->onCcd(mmX, mmY);
	if (ccdnr == Xmm::CCD_NONE)
//{errstr<<"column/row=("<<columnID<<","<<rowID<<")->CAMCOORD2=("<<mmX<<","<<mmY<<") not on any CCD."<<Sas::message(Msg::App,Msg::Always);
	  continue; // while (shapeScanner...)
//}
	int8 int8ccdnr = (int8)ccdnr;
	CalInt8Vector ccdnrVector(&int8ccdnr, 1u);

	// Check for bad pixels.
	XmmPixelList rawXY = detXY.toChipCoord().toPixCoord1(ccdnrVector);
//errstr << "a" << Sas::message(Msg::App,Msg::Always);
/*
errstr << "column "<<columnID << " row " << rowID << " mmX " << mmX << " mmY " << mmY << " mmZ " << mmZ << Sas::message(Msg::App,Msg::Always);
errstr<<"ccd "<<ccdnr<<"("<<(rawXY.x())[0] << "," << (rawXY.y())[0] << ")" << Sas::message(Msg::App,Msg::Always);
errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr<<"bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
errstr << "bool** =";
for (unsigned i=0u; i < _nCcds; i++)
  errstr << " " << _goodPixelMaps[i];
errstr<< Sas::message(Msg::App,Msg::Always);
errstr << "_goodPixelCounts[0] ="<< _goodPixelCounts[0] << Sas::message(Msg::App,Msg::Always);
errstr << "bool*  ="<< _goodPixelMaps[0][0] << Sas::message(Msg::App,Msg::Always);
errstr
<< _goodPixelMaps[0][0][0]
<< _goodPixelMaps[4][301][301]
<< _goodPixelMaps[5][299][300]
<< _goodPixelMaps[ccdnr - EMOScam::CCD_FIRST]
			[ (rawXY.x())[0] - _rawXmin ]
			[ (rawXY.y())[0] - _rawYmin ]
<< Sas::message(Msg::App,Msg::Always);
*/
	if (!_goodPixelMaps[ccdnr - EMOScam::CCD_FIRST]
			[ (rawXY.x())[0] - _rawXmin ]
			[ (rawXY.y())[0] - _rawYmin ])
//{ errstr << "b" << Sas::message(Msg::App,Msg::Always);
//{errstr<<"("<<columnID<<","<<rowID<<")->("<<ccdnr<<","<<(rawXY.x())[0]<<","<<(rawXY.y())[0]<<") bad."<<Sas::message(Msg::App,Msg::Always);
	  continue; // while (shapeScanner...)
//}
//else { errstr << "c" << Sas::message(Msg::App,Msg::Always);
//}
//errstr << "d" << " " << ccdnr << Sas::message(Msg::App,Msg::Always);
//errstr<<"("<<columnID<<","<<rowID<<")->("<<ccdnr<<","<<(rawXY.x())[0]<<","<<(rawXY.y())[0]<<") accepted."<<Sas::message(Msg::App,Msg::Always);

	// Passed all tests.  Remember the good pixel.
	goodPixelCount++;
      } // for (int rowID = span.min; rowID <= span.max; rowID++)
//errstr << "e" << Sas::message(Msg::App,Msg::Always);
    } // for (list<scanningLib::Span>::iterator spanIterator = spanlist.begin();
//errstr << "f" << Sas::message(Msg::App,Msg::Always);
  } // while shapeScanner.next(columnID, spanlist)
  calServer.releaseAtom(gds);
//errstr << "g" << Sas::message(Msg::App,Msg::Always);
  return goodPixelCount;
} // unsigned long PixelCounter::filterRegion(scanningLib::PolygonScanner shapeScanner)

//------------------------------------------------------------------------------
  // Null the flags of pixels unselected.
  // (Obtain an iterator from the array.  Look at rgslib/lib/ShapeScanner.h)

//------------------------------------------------------------------------------
/** Just a trivial class to produce a vector<Vertex>
	for the corners of a CCD in PIXCOORD1 (RAWX/Y). */
vector<scanningLib::Vertex> PixelCounter::CCDcorners() {
//errstr << "CCDcorners " << "(1)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  vector<scanningLib::Vertex> cornerVertices(4);
/*
  cornerVertices[0] = scanningLib::Vertex(_rawXmin-0.5, _rawYmin-0.5);
  cornerVertices[1] = scanningLib::Vertex(_rawXmax+0.5, _rawYmin-0.5);
  cornerVertices[2] = scanningLib::Vertex(_rawXmax+0.5, _rawYmax+0.5);
  cornerVertices[3] = scanningLib::Vertex(_rawXmin-0.5, _rawYmax+0.5);
*/
//errstr << "CCDcorners " << "(2)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  cornerVertices[0] = scanningLib::Vertex(_rawXmin-1, _rawYmin-1);
//errstr << "CCDcorners " << "(3)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  cornerVertices[1] = scanningLib::Vertex(_rawXmax+1, _rawYmin-1);
//errstr << "CCDcorners " << "(4)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  cornerVertices[2] = scanningLib::Vertex(_rawXmax+1, _rawYmax+1);
//errstr << "CCDcorners " << "(5)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  cornerVertices[3] = scanningLib::Vertex(_rawXmin-1, _rawYmax+1);
//errstr << "CCDcorners " << "(6)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  return cornerVertices;
} // vector<scanningLib::Vertex> CCDcorners()

//------------------------------------------------------------------------------
/** A not-so-trivial class to produce a list<Vertex>
	for the corners of the named CCD in CAMCOORD2 (DETX/Y),
	for feeding into PolygonScanner::(list<scanningLib::Vertex>). */
list<scanningLib::Vertex> PixelCounter::CCDcorners(Ccd ccdnr) {
  vector<scanningLib::Vertex> cornerVertexVector = CCDcorners();
//errstr << "CCDcorners("<<ccdnr<<") " << "(1)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
//  GeometryDataServer * gds = 0; gds = calServer.getAtom(gds);
//errstr << "CCDcorners("<<ccdnr<<") " << "(1.5)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  long vectorLength = cornerVertexVector.size();
  int16 * rawX = new int16[vectorLength];
  int16 * rawY = new int16[vectorLength];
//  int8 * ccdnrArray = new int8[vectorLength];
  for (int i=0; i < vectorLength; i++) {
    rawX[i] = (int16)(cornerVertexVector[i].x);
    rawY[i] = (int16)(cornerVertexVector[i].y);
//    ccdnrArray[i] = (int8)ccdnr;
  } // for (int i=0; i < vectorLength; i++)
  CalInt16Vector rawXvector(rawX, vectorLength);
  CalInt16Vector rawYvector(rawY, vectorLength);
//  CalInt8Vector ccdnrVector(ccdnrArray, vectorLength);

// PIXCOORD1 |-> CAMCOORD2
//errstr << "CCDcorners("<<ccdnr<<") " << "(2)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  calServer.getState().ccd()->set(ccdnr);
	// Must set ccdnr before calling calServer.getAtom(GeometryDataServer)
	//  or else coordinate conversion will be bogus.
//errstr << "CCDcorners("<<ccdnr<<") " << "(3)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  GeometryDataServer * gds = 0; gds = calServer.getAtom(gds);
//errstr << "CCDcorners("<<ccdnr<<") " << "(4)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  XmmCartesianPointList cornerCamCoord2List
	= gds->pixCoord1List(rawXvector, rawYvector)
//	.toChipCoord(&ccdnrVector)
	.toChipCoord()
	.toCamCoord1()
	.toCamCoord2();
/*
  XmmPixelList cornerPixCoord1List = gds->pixCoord1List(rawXvector, rawYvector);
for (int i=0; i < vectorLength; i++) errstr << " " << rawX[i] << " " << rawY[i];
errstr << Sas::message(Msg::App,Msg::Always);
//  XmmPixelList cornerChipCoordList = cornerPixCoord1List.toChipCoord(&ccdnrVector);
  XmmPixelList cornerChipCoordList = cornerPixCoord1List.toChipCoord();
for (int i=0; i < vectorLength; i++) errstr << " " << rawX[i] << " " << rawY[i];
errstr << Sas::message(Msg::App,Msg::Always);
  XmmCartesianPointList cornerCamCoord1List = cornerChipCoordList.toCamCoord1();
for (int i=0; i < vectorLength; i++) errstr << " " << rawX[i] << " " << rawY[i];
errstr << Sas::message(Msg::App,Msg::Always);
  XmmCartesianPointList cornerCamCoord2List = cornerCamCoord1List.toCamCoord2();
*/
//errstr << "CCDcorners("<<ccdnr<<") " << "(5)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
errstr << "corner rawXY: ";
for (int i=0; i < vectorLength; i++) errstr << " " << rawX[i] << " " << rawY[i];
errstr << Sas::message(Msg::App,Msg::Always);
  calServer.releaseAtom(gds);
/*
for (unsigned i=0; i < 4; i++) {
errstr
<<" "<<cornerVertexVector[i].x
<<" "<<cornerVertexVector[i].y
<<" "<<rawX[i]
<<" "<<rawY[i]
<<" "<<rawXvector[i]
<<" "<<rawYvector[i]
<<" "<<(cornerPixCoord1List.x())[i]
<<" "<<(cornerPixCoord1List.y())[i]
<<" "<<(cornerChipCoordList.x())[i]
<<" "<<(cornerChipCoordList.y())[i]
<<" "<<(cornerCamCoord1List.x())[i]
<<" "<<(cornerCamCoord1List.y())[i]
<<" "<<(cornerCamCoord2List.x())[i]
<<" "<<(cornerCamCoord2List.y())[i]
<<Sas::message(Msg::App,Msg::Always);
}
*/
//errstr << "CCDcorners("<<ccdnr<<") " << "(6)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  delete [] rawX; rawX = 0;
  delete [] rawY; rawY = 0;
//  delete [] ccdnrArray; ccdnrArray = 0;

// CAMCOORD2 |-> DETX/Y
  MiscDataServer * mds = 0; mds = calServer.getAtom(mds);
  double mmX_to_detX =	  mds->get( "PLATE_SCALE_X") * 20.0
			/ mds->get("MM_PER_PIXEL_X");
  double mmY_to_detY =	  mds->get( "PLATE_SCALE_Y") * 20.0
			/ mds->get("MM_PER_PIXEL_Y");
	// Reference: packages/ccftools/coordconv/src/camtochip.cc
  calServer.releaseAtom(mds);

// The headache here is that conversion between PIXCOORD1 and CAMCOORD2
// requires (x,y) to be in containers defined in DAL (specifically, the
// XmmCartesianPointLists in GeometryDataServer.h).  On the other hand,
// we use the PolygonScanner class in rgslib to iterate through all pixels
// within a CCD.  PolygonScanner requires (x,y) to be in list<Vertex>.

//errstr << "CCDcorners("<<ccdnr<<") " << "(7)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  CalReal64Vector mmX = cornerCamCoord2List.x();
  CalReal64Vector mmY = cornerCamCoord2List.y();
//errstr << "CCDcorners("<<ccdnr<<") " << "(8)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  list<scanningLib::Vertex> cornerVertexList;
//errstr << "CCDcorners("<<ccdnr<<") " << "(9)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  for (unsigned i=0u; i < mmX.size(); i++)
{errstr<<"ccd "<<ccdnr<<" corner "<<i<<" mmXY: "<<mmX[i]<<","<<mmY[i]<<" detXY: "<<mmX[i]*mmX_to_detX<<","<<mmY[i]*mmY_to_detY<<Sas::message(Msg::App,Msg::Always);
    cornerVertexList.push_back( scanningLib::Vertex(	mmX[i] * mmX_to_detX,
							mmY[i] * mmY_to_detY) );
}

//errstr << "CCDcorners("<<ccdnr<<") " << "(10)" << Sas::message(Msg::App,Msg::Always);
//errstr << "bool***="<< _goodPixelMaps << Sas::message(Msg::App,Msg::Always);
//errstr << "bool** ="<< _goodPixelMaps[0] << Sas::message(Msg::App,Msg::Always);
  return cornerVertexList;
} // list<scanningLib::Vertex> PixelCounter::CCDcorners(Ccd ccdnr)

// EOF

/*
$Log: PixelCounter.cc,v $
Revision 0.4  2008/10/25 23:02:13  cmhchen
First revision with a working PixelCounter and task goodpixelcount.
New class AnomalousStateFinder and task emtaganomalous.
PixelCounter.cc: Major bug fix.

Revision 0.3  2008/10/10 10:08:36  cmhchen
This revision accepts a user-input selectlib-style expression for filtering,
in addition to discounting bad pixels based on the BADPIX0n extensions.
It counts in integral DETX/Y coordinates, covering all pixels on any CCD.
Works in principle, but looping through all 7 x 12000^2 DETX/Y takes forever.
Bug: PixelCounter::_goodPixelMaps is never initialised; see README.

Revision 0.2  2008/09/09 17:16:26  cmhchen
Restructured code from previous version to formal class PixelCounter.
goodpixelcount.cc modified as the new PixelCounter.cc .
There should be no change in functionality whatsoever.

Revision 0.1  2008/09/05 18:02:37  cmhchen
First sandbox version that does bad pixel filtering based on BADPIXnn.
Pixel counts are written to file header in PIXCOORD1 (RAWX/Y).

*/
