/**************************************************************
*
* printstat.cc	XMM Science Analysis System (SAS)
*
* For testing the AnomalousStateFinder class of routines.
*
* (c) 2008 ESA - Written by Hubert Chen, CEA Saclay
*
* $Id: printstat.cc,v 0.5 2008/10/28 17:07:34 cmhchen Exp $
*
**************************************************************/

// C++ global:
#include "printstat.h"
	// automatically generated; main() needs this.
// Implementation specific:
#include "AnomalousStateFinder.h"

/** Read a list of event file filename strings from STDIN,
    print one line per filename containing:
expID, and for each CCD, soft-band count, hard-band count, rate-band count, GTI sum, H, R, isAnomalous?
    That is, 1 + 7 x 7 = 50 columns per file.
    Files with no STDGTI0n extensions are skipped. */
/** Given the filename of a MOS merged event list (on the command line),
    decide if any CCD is in an anomalous state, and write the result
    as attributes to the input file header.
    This is the SAS task that gives access to the AnomalousStateFinder class. */
void printstat() {
  string filename;
  while ( getline(cin, filename) ) {
    DataSet * dataSetPtr = dataSetServer->open(filename, Dal::Read);
	// File to be opened to read EXP_ID and check for STDGTI0n extensions.
    bool GTIcomplete = true;
    for (unsigned i=1; i <= 7; i++) {
      ostringstream ccdnrStrStream;
      ccdnrStrStream << i;
      if (! dataSetPtr->hasBlock("STDGTI0"+ccdnrStrStream.str()) )
	GTIcomplete = false;
    } // for (unsigned i=1; i <= 7; i++) {
    if (!GTIcomplete) continue; // while ( getline(cin, filename) )

    cout << dataSetPtr->attribute("EXP_ID")->asString();
    AnomalousStateFinder asf(dataSetPtr);
    for (Ccd ccdnr = EMOScam::CCD_FIRST; ccdnr <= EMOScam::CCD_LAST; ccdnr++)
      cout
	<< "\t" << asf.getCount(ccdnr, AnomalousStateFinder::CCDdata::SOFTBAND)
	<< "\t" << asf.getCount(ccdnr, AnomalousStateFinder::CCDdata::HARDBAND)
	<< "\t" << asf.getCount(ccdnr, AnomalousStateFinder::CCDdata::RATEBAND)
	<< "\t" << asf.getGTIsum(ccdnr)
	<< "\t" << asf.getH(ccdnr)
	<< "\t" << asf.getR(ccdnr)
	<< "\t" << (asf.isAnomalous(ccdnr)? 1 : 0);
    cout << endl;
    dataSetServer->close(dataSetPtr);
  } // while ( getline(cin, filename) )
} // void printstat()

/*
$Log: printstat.cc,v $
Revision 0.5  2008/10/28 17:07:34  cmhchen
Initial revision.

*/
